// A. Sheaff 02/24/05

// Image class - provides basic image routines for gray scale images.

#ifndef __IMAGE_CC__
#define __IMAGE_CC__

#include <sys/types.h>

#include "xdisplayimage.hh"

#define Pixel u_int8_t


class Image
{
	protected:
		int w;		// Width
		int h;		// Height
		Pixel* d;	// Image data

		static XDisplayImage xdi;

		inline int index(int x, int y) const;		// Calculate the index into the data array

	public:
		// Constructors
		Image(void) : w(0), h(0), d(NULL) {};	// From nothing
		Image(int wd, int ht);		// From width and height
		Image(int wd, int ht, Pixel co);	// Of a particular color
		Image(const Image& i);	// From another image
		Image(const char * filename, const int wd, const int ht);	// From a file, one byte per pixel
		~Image(void);	// Destructor

		inline int width(void) const { return w; }	// Get the width
		inline int height(void) const { return h; }	// Get the height
		inline Pixel* data(void) const { return d; }	// Get a pointer to the data
		int size(void);	// Return size of image

		Pixel getpixel(const int x, const int y) const;	// Get a pixel value
		Pixel setpixel(const int x, const int y, const Pixel p);	//Set a pixel value

		Image& operator=(const Image& i);	// Assignment
		Image operator~(void);		// Logical not of an image

		Image& operator+=(const Pixel c);	// Add a constant
		Image& operator-=(const Pixel c);	// Subract a constant
		Image& operator|=(const Pixel c);	// Or a constant
		Image& operator&=(const Pixel c);	// And a constant
		Image& operator^=(const Pixel c);	// Xor a constant

		void display(void);	// Display the image
		void refresh(void);	// Refresh display with text and lines
		void pause(void);	// Wait for a keypress
		void title(char * t);	// Set the window title
		void drawline(int x1, int y1, int x2, int y2);	// Draw a line
		void setlinewidth(int w);	// Set the line width
		void setcolor(unsigned long c);	// Set the drawing color
		void drawtext(int x, int y, char * string);	// Draw text

		int threshold(Pixel lt, Pixel ut);	// Threshold an image

		int writeimage(char * filename);	// Write file out in gray format
									// Use convert -depth 8 -size WxH to convert
};

#endif
