
//  Copyright 1992, Massachusetts Institute of Technology. All Rights Reserved.

// --------------------------------------------------------------------
// Permission to use, copy, or modify this software and its documentation
// for educational and research purposes only and without fee is hereby
// granted, provided that this copyright notice appear on all copies and
// supporting documentation.  For any other uses of this software, in
// original or modified form, including but not limited to distribution
// in whole or in part, specific prior permission must be obtained from
// M.I.T. and the authors.  These programs shall not be used, rewritten,
// or adapted as the basis of a commercial software or hardware product
// without first obtaining appropriate licenses from M.I.T.  M.I.T. makes
// no representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.

// ---------------------------------------------------------------------

//   Author: Matthew Turk
//   Date: 3/22/91
//   Modified:  Thad Starner 10/6/91

//   Hacked by A Sheaff AI MIT 1/20/00
  
//	Display an image under X
//	H H Atkins - AI Lab MIT 01/12/00
//	A Sheaff - AI Lab MIT 01/12/00

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
}

#include "xdisplayimage.hh"

// Create initial windows with image and title.  Any new images cannot exceed init dims.
XDisplayImage::XDisplayImage(void) : Dpy(NULL), Pseudocolor_Visual(NULL), VList(NULL), Img(NULL), Buffer(NULL)
{
	Depth=3;
	Width=480;
	Height=640;
	create_window(Buffer, "XDisplayImage");
}

// Create initial windows with image and title.  Any new images cannot exceed init dims.
XDisplayImage::XDisplayImage(int width, int height, char * title) : Dpy(NULL), Pseudocolor_Visual(NULL), VList(NULL), Img(NULL), Buffer(NULL)
{
	Depth=3;
	Width=width;
	Height=height;
	create_window(Buffer, title);
}

// Clean up
XDisplayImage::~XDisplayImage(void)
{
	XShmDetach(Dpy,&XShMSI);
	XDestroyImage(Img);
	shmdt(Buffer);
	XDestroyWindow(Dpy,Win);
	shmctl(XShMSI.shmid,IPC_RMID,NULL);
}

// Input pixel [0,B,G,R] which is an int.
// Initialize the display
int XDisplayImage::create_window(unsigned char * buffer, char * title)
{
	int num_matched;
	XGCValues xgcv;
	XSetWindowAttributes xswa;
	unsigned long foreground, background;
	XSizeHints sizehints;
	XVisualInfo __template;
	Screen * s;

	// Open a connection
	if ((Dpy = XOpenDisplay(NULL)) == NULL) {
		fprintf(stderr, "Unable to open display: %s\n", XDisplayName(NULL));
		exit(1);
	}
	// Get characteristics
	Scrn = DefaultScreen(Dpy);
	s=ScreenOfDisplay(Dpy,Scrn);
	if ((Depth<0) || (Depth > DefaultDepth(Dpy,Scrn)/8)) {
		Depth=3;
	}
	if ((Width<0) || (Width > (WidthOfScreen(s)))) {
		Width=WidthOfScreen(s);
	}
	if ((Height<0) || (Height > (HeightOfScreen(s)))) {
		Height=(HeightOfScreen(s));
	}

	Root = RootWindow(Dpy, Scrn);
	Screen_Depth = (unsigned int)DefaultDepth(Dpy,Scrn);
	__template.depth = Screen_Depth;
	VList = XGetVisualInfo(Dpy, VisualDepthMask , &__template, &num_matched);
	if (num_matched == 0) {
		fprintf (stderr,"can't match visual\n");
		exit(1);
	}
	Pseudocolor_Visual = VList[0].visual;
	xswa.event_mask = (ExposureMask | StructureNotifyMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask);
	// Create a window and set properties
	Win = XCreateWindow(Dpy, Root, 0, 0, Width, Height, 1, DefaultDepth(Dpy, Scrn), InputOutput,DefaultVisual(Dpy, Scrn), CWEventMask | CWBorderPixel | CWBackPixel, &xswa);
	sizehints.flags = 0;
	XSetStandardProperties(Dpy, Win, title, "XWindow", None, 0, 0, &sizehints);
	// Our image (and lines and text) is contained in a graphics context
	GrCon = XCreateGC(Dpy, Win, 0, &xgcv);
	foreground = WhitePixel(Dpy, Scrn);
	background = BlackPixel(Dpy, Scrn);          
	//XSetErrorHandler (error_hook);
	// Display the window
	XMapWindow(Dpy, Win);
	XClearArea(Dpy, Win, 0, 0, 0, 0, 0);

	// Create a shared memory area for images and set properties
	Img = XShmCreateImage(Dpy, Pseudocolor_Visual, Screen_Depth, ZPixmap, NULL, &XShMSI, Width, Height);
	XShMSI.shmid=shmget(IPC_PRIVATE,Img->bytes_per_line*Img->height, IPC_CREAT|0777);
	if (XShMSI.shmid==-1) {
		perror("shmget()");
		exit(0);
	}
	XShMSI.shmaddr=(char *)shmat(XShMSI.shmid,0,0);
	Img->data=XShMSI.shmaddr;
	Buffer=(unsigned char *)XShMSI.shmaddr;
	Buf_Size=Img->bytes_per_line*Img->height;
	XShMSI.readOnly=False;
	XShmAttach(Dpy,&XShMSI);
	return(0);
}

// Call with new buffer to update display.
void XDisplayImage::paint(const unsigned char * buf, const int width, const int height)
{
	int i,j,twidth,theight;
	unsigned int *pixels=(unsigned int *)Buffer;

	if (width>Width) {
		twidth=Width;
	} else {
		twidth=width;
	}
	if (height>Height) {
		theight=Height;
	} else {
		theight=height;
	}
	if (buf) {
		for (i=0;i<theight;i++) {
			for (j=0;j<twidth;j++) {
				pixels[Width*i+j]=(buf[twidth*i+j]<<16)+(buf[twidth*i+j]<<8)+buf[twidth*i+j];
			}
		}
	}
	XShmPutImage(Dpy, Win, GrCon, Img, 0, 0, 0, 0, Width, Height,False);
	XSync(Dpy,False);
}

// Catches X Errors
int error_hook (Display * display, XErrorEvent * err)
{
	char msg[80];
	XGetErrorText (display, err->error_code, msg,80);
	printf ("X WINDOW ERROR: %s\n", msg);
	return(0);
}

// Set the window title
void XDisplayImage::title(char * t)
{
	XSizeHints sizehints;

	sizehints.flags = 0;
	XSetStandardProperties(Dpy, Win, t, "XWindow", None, 0, 0, &sizehints);
	return;
}

// Draw a line.  Calling update removes any drawing
void XDisplayImage::drawline(int x1, int y1, int x2, int y2)
{
	XDrawLine(Dpy,Win,GrCon,x1,y1,x2,y2);
	return;
}

// Set the drawing line width
void XDisplayImage::setlinewidth(int w)
{
	XSetLineAttributes(Dpy,GrCon,w,LineSolid,CapNotLast,JoinMiter);
	return;
}

// Set the drawing forground color
void XDisplayImage::setcolor(unsigned long c)
{
	XSetForeground(Dpy,GrCon,c);
	return;
}

// Accept keyboard input while the window is in focus
int XDisplayImage::waitforevent(void)
{
	XEvent event_return;

	XNextEvent(Dpy,&event_return);
	while (event_return.type!=KeyPress) {
		XNextEvent(Dpy,&event_return);
		if (event_return.type==Expose) {
			XShmPutImage(Dpy, Win, GrCon, Img, 0, 0, 0, 0, Width, Height,False);
			XSync(Dpy,False);
		}
	}

	if (event_return.type==KeyPress) {
		if (XKeycodeToKeysym(Dpy,event_return.xkey.keycode,0)=='q') {
			exit(0);
		}
		if (XKeycodeToKeysym(Dpy,event_return.xkey.keycode,0)=='p') {
			return (-1);
		}
	}
	return (1);
}

// Draw some text
int XDisplayImage::drawtext(int x, int y, char * string)
{
	XDrawString(Dpy,Win,GrCon,x,y,string,strlen(string));
	return 0;
}

void XDisplayImage::refresh(void)
{
	XSync(Dpy,False);
	return;
}
